/*
   Copyright 2005 Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
    * Neither the name of Red Hat, Inc., nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "../config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <glob.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <syslog.h>
#include <dbus/dbus.h>
#include "handlers.h"
#include "util.h"

#ident "$Id: launch.c,v 1.3 2005/06/27 23:30:21 nalin Exp $"

static int
run(const char *service, const char *path)
{
	printf("About to start %s (\"%s\").\n", service, path);
	/* FIXME */
	return 0;
}

/* Given a service name, scan the .desktop files and locate the binary which
 * should implement that service. */
static int
launch(const char *service)
{
	glob_t globbed;
	char path[PATH_MAX + 1], names[8192], buf[8192], *p;
	struct {
		const char key[7];
		size_t keylen;
		char *buf;
		size_t bufsize;
	} vars[] = {
		{"Names=", sizeof("Names=") - 1, names, sizeof(names)},
		{"Name=", sizeof("Name=") - 1, names, sizeof(names)},
		{"Exec=", sizeof("Exec=") - 1, path, sizeof(path)},
	};
	FILE *fp;
	unsigned i, j, span;
	const char *section = "[D-BUS Service]";
	dbus_bool_t found, in_section, dupe;

	i = glob(DATADIR "/" DBUS_PACKAGE "/services/*.desktop",
		 0, NULL, &globbed);
	if (i != 0) {
		return -1;
	}

	found = dupe = FALSE;
	for (i = 0; i < globbed.gl_pathc; i++) {
		in_section = FALSE;
		fp = fopen(globbed.gl_pathv[i], "r");
		if (fp != NULL) {
			strcpy(path, "");
			strcpy(names, "");
			while (fgets(buf, sizeof(buf), fp) != NULL) {
				/* If it looks like a section start, check if
				 * it's the section we need to heed. */
				if (buf[0] == '[') {
					in_section = (strncmp(buf, section,
							      strlen(section)) == 0) &&
						     (strchr("\r\n",
							     buf[strlen(section)]) != NULL);
					continue;
				}
				if (!in_section) {
					continue;
				}
				/* Parse out variables about which we care. */
				for (j = 0;
				     j < sizeof(vars) / sizeof(vars[0]);
				     j++) {
					if (vars[j].keylen <= 0) {
						vars[j].keylen =
							strlen(vars[j].key);
					}
					if (strncmp(buf, vars[j].key,
						    vars[j].keylen) == 0) {
						snprintf(vars[j].buf,
							 vars[j].bufsize,
							 "%s",
							 buf + vars[j].keylen);
						/* Snip of trailing CR/LFs. */
						span = strcspn(vars[j].buf,
							       "\r\n");
						if (span > 0) {
							vars[j].buf[span] = '\0';
						}
						break;
					}
				}
			}
			/* If we got a path and names, check if we have a match
			 * on the service name. */
			if ((strlen(path) > 0) && (strlen(names) > 0)) {
				p = names;
				while (*p == ';') {
					p++;
				}
				while (*p != '\0') {
					span = strcspn(p, ";\r\n");
					if (span > 0) {
						if (strncmp(service, p,
							    span) == 0) {
							if (found) {
								/* Duplicate! */
								dupe = TRUE;
							}
							found = TRUE;
							break;
						}
					}
					p += span;
					p += strspn(p, ";\r\n");
				}
			}
			fclose(fp);
		}
	}
	globfree(&globbed);
	/* If we got a match, run it. */
	if (found && !dupe) {
		return run(service, path);
	}
	return HANDLER_FAILURE;
}

int
main(int argc, char **argv)
{
	char arg[LINE_MAX];
	int i;
	openlog(PACKAGE "-dbus-launch", LOG_PID, LOG_DAEMON);
	if (fgets(arg, sizeof(arg), stdin) != NULL) {
		i = strcspn(arg, "\r\n");
		arg[i] = '\0';
		if (strlen(arg) > 0) {
			i = launch(arg);
			closelog();
			return i;
		}
	}
	closelog();
	return HANDLER_INVALID_INVOCATION;
}
